#!/usr/bin/env python3
import os.path

import paramiko
import argparse
import multiprocessing as mp
import time
import sys
import signal
import configparser
import uuid
import re

LINES_IN_STATUS = (3, 4, 5)  # master has 3 lines, slave has 4
STAT_NO_REPLICA = 'Status: Not initialized yet'  # disable setting controller to master if it reports that in status
STAT_IN_REPLICA = 'Status: Initialized'
STAT_REMOTE_SEEN = 'state: active'
STAT_IS_MASTER = 'Current node status: master'
STAT_IS_SLAVE = 'Current node status: slave'

SSH_LOOP_TIMEOUT = 300
LOOP_TIMEOUT = 10  # seconds between last cmd completed and status request
T_OVERHEAD = 1
EXIT_CMD = ['EXIT']
# [cmd, max timeout to wait in worker, debug_level to display stdout]; deadline = timeout+1sec
STATUS_CMD = ['/usr/local/sbin/veil-controller -w', 30, 10]
SET_ROLE_MASTER_CMD = ['/usr/local/sbin/veil-controller -r master', 240, 1]
NO_CONN_MUTE_AFTER = 10  # decrements every sequential unsuccessful connection; after this value no messages to log
REPORTS_PERIOD = 3600  # periodically print servers status to log (seconds)
CONFIRM_2SIMILAR_TIMEOUT = 30
MASTER_FAILURE_DURATION = 30  # how long master can be offline before slave will be switched to master
MAX_ID_LEN = 50
SWITCH_FLAG = '/tmp/role-switch'
DEBUG = 0


def ssh_process(ip, user, pipe: mp.Pipe(), timeout, password=None, verbose=True, pkey=None, debug=False):
    if ':' in ip:
        ip, port = ip.split(':', 1)
    else:
        port = 22

    signal.signal(signal.SIGTERM, sigterm_handler)
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    try:
        if password:
            ssh.connect(hostname=ip, port=port, username=user, password=password, look_for_keys=False,
                        allow_agent=False,
                        timeout=10)
        else:
            pk = paramiko.RSAKey.from_private_key_file(pkey)
            ssh.connect(hostname=ip, port=port, username=user, pkey=pk, timeout=10)

    except Exception as e:
        if verbose:
            print(f'Exception while establishing SSH session to {ip} e={e}')
        pipe.close()
        exit(2)

    while True:
        try:
            if not pipe.poll(timeout):
                print(f'{ip} process got no commands from parent in {timeout} seconds. Exiting...')
                ssh.close()
                exit(0)

            try:
                rcv_data = pipe.recv()  # [str cmd, int timeout]
            except EOFError:
                print(f'{ip} process got EOF (parent closed pipe). Exiting...')
                exit(1)

            if rcv_data[0] == 'EXIT':
                print(f'{ip} Exiting by parent command...')
                ssh.close()
                break
            else:
                start = time.monotonic()
                stdin, stdout, stderr = ssh.exec_command(rcv_data[0], timeout=rcv_data[1])
                stdin.close()
                channel = stdout.channel
                _out, _err = '', ''
                while channel.recv_ready() or channel.recv_stderr_ready() or not channel.closed:
                    _chunk_out, _chunk_err = '', ''
                    while channel.recv_ready() or channel.recv_stderr_ready():
                        if channel.recv_ready():
                            _chunk_out += channel.recv(512).decode()
                        if channel.recv_stderr_ready():
                            _chunk_err += channel.recv_stderr(512).decode()
                    if _chunk_out:
                        if len(rcv_data) == 3 and debug // rcv_data[2] > 1:
                            log('\n'.join([f'{ip} >> {_line}' for _line in _chunk_out.splitlines()]))
                        _out += _chunk_out
                    if _chunk_err:
                        if len(rcv_data) == 3 and debug // rcv_data[2] > 1:
                            log('\n'.join([f'{ip} E> {_line}' for _line in _chunk_out.splitlines()]))
                        _err += _chunk_err
                    time.sleep(0.1)
                if len(rcv_data) == 3 and debug // rcv_data[2] > 0:
                    log(f'Command "{rcv_data[0]}" executed with return code {channel.recv_exit_status()} '
                        f'in {round(time.monotonic() - start, 1)} sec.')
            pipe.send([_out, _err])

        except KeyboardInterrupt:  # kill with SIGINT
            print(f'{ip} Exiting by SIGINT...')
            ssh.close()
            break

        except paramiko.ssh_exception.SSHException as se:
            pipe.send(['', 'ERROR'])
            if verbose:
                print(f"{ip} SSH exception, e={se}. Exiting...")
            break

    pipe.close()
    exit(0)


def sigterm_handler(sig, stack_frame):
    print('Got SIGTERM, exiting ...')
    raise SystemExit


def log(string):
    print(string, flush=True)


def main():
    parser = argparse.ArgumentParser(description='Checks for availability of remote master controller and switches'
                                                 ' slave controller to master if current master is unavailable.')
    parser.add_argument('--check', action='store_true', help='Check config file for presence of required and those '
                                                             'specified, optional parameters '
                                                             '(consult man file for parameters description).')
    parser.add_argument('--configfile', default='/etc/default/controller-witness',
                        help='Custom config file location (default is /etc/default/controller-witness)')
    args = parser.parse_args()
    cfg = configparser.ConfigParser()
    try:
        cfg.read(args.configfile)
    except configparser.ParsingError as e:
        log(f'Config file {args.configfile} parsing error. e={e}')
        sys.exit(1)

    _id = cfg.get('default', 'id', fallback='')
    match = re.fullmatch(r'[0-9a-zA-Z_\-]+', _id)
    if not (_id or log('No ID in config file')) or \
            not (match or log('Wrong characters in ID! Allowed characters set: 0-9a-zA-Z_-')) or \
            (len(_id) > MAX_ID_LEN and not log('ID max length exceeded ({} > {})'.format(len(_id), MAX_ID_LEN))):
        if args.check:  # since it is called in postinst it is called from root so we are able to write cfg file here
            new_id = str(uuid.uuid4())
            log('New ID created: {}'.format(new_id))
            if 'default' in cfg.sections():
                cfg['default']['id'] = new_id
            else:
                cfg['default'] = dict(id=new_id)

            with open(args.configfile, 'w') as w_cfg:
                cfg.write(w_cfg)
        else:
            log('Wrong instance ID. Consult man about instance ID requirements. Exiting...')
            exit(1)

    if not ('default' in cfg.sections()):
        log('[default] section is absent in config file!')
        exit(1)
    if not cfg['default'].get('server1') or not cfg['default'].get('server2') \
        or not (cfg['default'].get('username') or (cfg['default'].get('username1')
                                                   and cfg['default'].get('username2'))):
        log('All parameters "server1", "server2", "username1", "username2" or '
            '"username" common for both servers must be present in [default] section!')
        exit(1)

    comm_user, comm_pwd = cfg['default'].get('username'), cfg['default'].get('password')
    comm_pkey = cfg['default'].get('pkey')

    if comm_user:
        cfg['default']['username1'] = cfg['default']['username2'] = comm_user
    if comm_pwd:
        cfg['default']['password1'] = cfg['default']['password2'] = comm_pwd
    elif comm_pkey:
        cfg['default']['pkey1'] = cfg['default']['pkey2'] = comm_pkey

    ips = [cfg['default']['server1'], cfg['default']['server2']]
    if ips[0] == ips[1]:
        log('IP addresses and/or ports must be different!')
        exit(1)

    ip_data = {ips[0]: dict(user=cfg['default']['username1'],
                            password=cfg['default'].get('password1'),
                            pkey=cfg['default'].get('pkey1')),
               ips[1]: dict(user=cfg['default']['username2'],
                            password=cfg['default'].get('password2'),
                            pkey=cfg['default'].get('pkey2'))}
    warn = ''
    for ip, ip_d in ip_data.items():
        if not (ip_d['password'] or ip_d['pkey']):
            warn += f'No password or pkey file is given for {ip}!\n'
    if warn:
        exit(1)

    try:
        SET_ROLE_MASTER_CMD[1] = cfg.getint('optional', 'set_master_timeout', fallback=SET_ROLE_MASTER_CMD[1])
        STATUS_CMD[1] = cfg.getint('optional', 'status_timeout', fallback=STATUS_CMD[1])
        loop_timeout = cfg.getint('optional', 'loop_timeout', fallback=LOOP_TIMEOUT)
        ssh_loop_timeout = cfg.getint('optional', 'ssh_loop_timeout', fallback=SSH_LOOP_TIMEOUT)
        no_conn_mute = cfg.getint('optional', 'no_conn_mute', fallback=NO_CONN_MUTE_AFTER)
        reports_period = cfg.getint('optional', 'reports_period', fallback=REPORTS_PERIOD)
        confirm_2similar_timeout = cfg.getint('optional', 'confirm_2similar_timeout', fallback=CONFIRM_2SIMILAR_TIMEOUT)
        master_failure_duration = cfg.getint('optional', 'master_failure_duration', fallback=MASTER_FAILURE_DURATION)
        switch_flag = cfg.get('optional', 'switch_flag', fallback=SWITCH_FLAG)
        debug = cfg.getint('optional', 'debug', fallback=DEBUG)
        print('debug is', debug)
    except ValueError as e:
        log(f'Wrong parameters on optional section, e={e}')
        exit(1)

    if args.check:
        exit(0)

    log(f'Starting controller-witness id={_id}')
    STATUS_CMD[0] += f' {_id}'

    uptime = time.monotonic()
    processes = {ip: dict() for ip in ips}
    deadline_2masters = 0
    deadline_2slaves = 0
    signal.signal(signal.SIGTERM, sigterm_handler)
    ignore_flag = False

    try:
        while True:
            for ind, ip in enumerate(ips):
                other_ip = ips[1 - ind]

                pipe, proc, deadline, is_master = processes[ip].get('pipe'), processes[ip].get('proc'), \
                    processes[ip].get('deadline'), processes[ip].get('is_master')

                # SSH sessions (re)establishing
                if not (proc and proc.is_alive()):  # no session yet exist or existing session failed
                    no_conn = min(0, processes[ip].get('conn_ok', 0))  # if was online before (>0) then 0 else negative
                    if proc:
                        if no_conn >= -no_conn_mute:
                            log(f"SSH session to {ip} exited with code {proc.exitcode}.")
                        if is_master > 0:
                            log(f'{ip} was master and lost connection')
                            if processes[other_ip]['proc'].is_alive():
                                processes[other_ip]['must_be_master'] = True  # obvious connect loss, no timeouts

                    if no_conn >= -no_conn_mute:
                        log(f'Starting SSH session to {ip}')
                    parent_conn, child_conn = mp.Pipe()
                    processes[ip] = dict(proc=mp.Process(target=ssh_process,
                                                         kwargs=dict(ip=ip,
                                                                     user=ip_data[ip]['user'],
                                                                     password=ip_data[ip]['password'],
                                                                     pipe=child_conn,
                                                                     timeout=ssh_loop_timeout,
                                                                     verbose=(no_conn > -no_conn_mute),
                                                                     pkey=ip_data[ip]['pkey'],
                                                                     debug=debug)),
                                         # Process class instance of SSH session worker
                                         pipe=parent_conn,  # Pipe handle to send cmds and get results from worker
                                         status='IDLE',  # detected command processing state
                                         in_replica=False,  # is this node a controller configured for replication
                                         is_master=-1,  # no state yet detected
                                         must_be_master=False,
                                         conn_ok=no_conn - 1,  # > 0  time when first reply from server received
                                         # <= 0 -number of failed sequential connect retries
                                         deadline=time.monotonic()  # when last cmd was completed (state IDLE)
                                         # or has to be completed (other)
                                         )
                    processes[ip]['proc'].start()
                    continue

                # main loop processing (0): role switch flag detecting and processing
                if os.path.isfile(switch_flag):
                    if ignore_flag:
                        try:
                            os.remove(switch_flag)
                        except OSError:
                            pass
                        else:
                            ignore_flag = False
                    else:
                        log('Switch flag detected.')
                        try:
                            os.remove(switch_flag)
                        except OSError as e:
                            log(f'Unable to remove switch flag. e={e}')
                            ignore_flag = True
                        if (processes[ip].get('conn_ok', 0) > 0 and processes[other_ip].get('conn_ok', 0) > 0
                            or log('Not all servers are connected. No switch by flag.')) and \
                                (deadline_2masters == 0 or log('2 masters state detected. No switching by flag.')) and \
                                (deadline_2slaves == 0 or log('2 slaves state detected. No switching by flag.')) and \
                                ((not (processes[ip]['must_be_master']
                                       or processes[other_ip]['must_be_master']
                                       or processes[ip]['status'] == 'SET_MASTER'
                                       or processes[other_ip]['status'] == 'SET_MASTER'))
                                 or log('Currently in switching state. No switching by flag.')) and \
                                ((processes[ip]['is_master'] > 0) is not (processes[other_ip]['is_master'] > 0)
                                 or log('2 servers in same state. No switching by flag.')) and \
                                (processes[ip]['is_master'] >= 0 and processes[other_ip]['is_master'] >= 0
                                 or log('One of servers is in unknown state. No switching by flag.')):
                            switch_to = ip if processes[other_ip]['is_master'] > 0 else other_ip
                            processes[switch_to]['must_be_master'] = True
                            log(f'Switching {switch_to} to role master by flag.')

                # main loop[1]: actions if needed
                if processes[ip]['status'] == 'IDLE':
                    if processes[ip]['must_be_master'] and processes[ip]['conn_ok'] > 0:
                        log(f'Setting controller {ip} role master')
                        try:
                            pipe.send(SET_ROLE_MASTER_CMD)
                        except BrokenPipeError:
                            log(f'{ip} Broken pipe. Trying to reconnect...')
                            processes[ip] = dict()
                            continue
                        processes[ip]['deadline'] = time.monotonic() + SET_ROLE_MASTER_CMD[1] + T_OVERHEAD
                        processes[ip]['status'] = 'SET_MASTER'
                        processes[ip]['must_be_master'] = False
                        processes[ip]['is_master'] = -1
                        # to avoid 'two slaves' situation in iteration next to cmd "set master" reply

                    elif deadline < time.monotonic() - loop_timeout:
                        try:
                            pipe.send(STATUS_CMD)
                        except BrokenPipeError:
                            log(f'{ip} Broken pipe. Trying to reconnect...')
                            processes[ip] = dict()
                            continue
                        processes[ip]['deadline'] = time.monotonic() + STATUS_CMD[1] + T_OVERHEAD
                        processes[ip]['status'] = 'W_STAT'

                # main loop processing (2): results of previous actions waiting and processing
                elif processes[ip]['status'] in ('W_STAT', 'SET_MASTER') and \
                        ((time.monotonic() > deadline) or pipe.poll()):

                    if not pipe.poll():
                        if is_master > 0:
                            processes[other_ip]['must_be_master'] = True  # change role cmd will be sent in next iter.
                        log(f'Session {ip} status request deadline exceeded')
                        processes[ip]['conn_ok'] = 0
                    else:
                        processes[ip]['deadline'] = time.monotonic()
                        try:
                            stdout, stderr = pipe.recv()  # here we receive list [stdout, stderr]
                        except ConnectionResetError:
                            log(f'ConnectionReset from {ip}')
                            processes[ip] = dict()
                            continue
                        else:
                            if processes[ip]['conn_ok'] <= 0:
                                processes[ip]['conn_ok'] = time.monotonic()
                                log(f'Established connection with {ip}')

                        if stderr == 'ERROR':
                            if is_master > 0:
                                processes[other_ip]['must_be_master'] = True
                            log(f'Session {ip} failed to send previous command')
                        else:
                            lines = stdout.splitlines()
                            # log(f'{ip} {processes[ip]["status"]}\n{lines}')

                            if processes[ip]['status'] == 'W_STAT' and len(lines) in LINES_IN_STATUS:  # parsing status

                                if lines[0].startswith(STAT_NO_REPLICA):
                                    log(f'{ip} controller not in replica')
                                    processes[ip]['in_replica'] = False
                                elif lines[0].startswith(STAT_IN_REPLICA):
                                    if not processes[ip]['in_replica']:
                                        log(f'{ip} is configured for replication')
                                        processes[ip]['in_replica'] = True

                                    if lines[1].startswith(STAT_IS_MASTER):
                                        deadline_2slaves = 0
                                        if is_master <= 0:  # -1 is "not yet known role from status", 0 is 'was slave'
                                            log(f'{ip} is master')
                                            processes[ip]['is_master'] = 1
                                        else:
                                            processes[ip]['is_master'] += 1

                                        if processes[other_ip]['is_master'] > 0 and \
                                                not (processes[ip]['status'] == 'SET_MASTER'
                                                     or processes[other_ip]['status'] == 'SET_MASTER'):
                                            # 2 masters (ex.: one has restored or catching switch roles via CLI moment )
                                            if deadline_2masters == 0:
                                                # we've just detected this state and setting timeout for situation to
                                                # self-resolve (if we got into a moment of CLI switching roles)
                                                deadline_2masters = time.monotonic() + confirm_2similar_timeout
                                                log(f'Two masters detected. Waiting {confirm_2similar_timeout} '
                                                    f'seconds to make decision.')
                                            elif deadline_2masters < time.monotonic():
                                                deadline_2masters = 0
                                                # setting as master the one who had been master longer
                                                log(f'Two masters state confirmed. {ip} is master for {is_master}'
                                                    f' iterations and {other_ip} is master'
                                                    f' for {processes[other_ip]["is_master"]} iterations')
                                                if is_master > processes[other_ip]['is_master']:
                                                    processes[ip]['must_be_master'] = True
                                                else:
                                                    processes[other_ip]['must_be_master'] = True
                                        else:
                                            deadline_2masters = 0

                                    elif lines[1].startswith(STAT_IS_SLAVE):
                                        deadline_2masters = 0
                                        if is_master != 0:  # -1 - not yet set; > 0 - was master before
                                            log(f'{ip} is slave')
                                        processes[ip]['is_master'] = 0
                                        if processes[other_ip]['is_master'] == 0 and \
                                                not (processes[ip]['status'] == 'SET_MASTER'
                                                     or processes[other_ip]['status'] == 'SET_MASTER'):
                                            if deadline_2slaves == 0:  # the same timeout as for 2 masters
                                                deadline_2slaves = time.monotonic() + confirm_2similar_timeout
                                                log(f'{ip} Two slaves detected. Waiting {confirm_2similar_timeout} '
                                                    f'seconds to make decision.')
                                            elif (deadline_2slaves < time.monotonic() and
                                                  not processes[other_ip]['must_be_master']):
                                                # if both controllers are slaves for 2similar_timeout or one
                                                # is slave and another is unavailable more than another timeout
                                                log(f'No master detected. Switching {ip} to master.')
                                                processes[ip]['must_be_master'] = True
                                                deadline_2slaves = 0
                                        elif processes[other_ip]['is_master'] < 0 and \
                                                processes[other_ip]['conn_ok'] < -master_failure_duration:
                                            # if one is slave and another is unavailable
                                            # for more than master_failure_timeout
                                            log(f'No master appeared in {master_failure_duration} seconds. '
                                                f'Switching the only living {ip} to master.')
                                            processes[ip]['must_be_master'] = True

                    processes[ip]['status'] = 'IDLE'

            if time.monotonic() - uptime > reports_period:
                uptime = time.monotonic()
                report = ""
                for ip in processes:
                    if processes[ip]['conn_ok'] > 0:
                        if processes[ip]['is_master'] > 0:
                            role = 'master'
                        elif processes[ip]['is_master'] == 0:
                            role = 'slave'
                        else:
                            role = 'unknown'
                        report += f'{ip}: {"" if processes[ip]["in_replica"] else "NOT "}in replica: {role}, online '
                        time_online = int(uptime - processes[ip]['conn_ok'])
                        timed = dict(day=time_online // 86400, hour=(time_online // 3600) % 24,
                                     minute=(time_online // 60) % 60, second=time_online % 60)
                        for k, v in timed.items():
                            report += f'{v} {k + ("s" if v != 1 else "")} ' if v or (k == 'second') else ""
                    else:
                        report += f'{ip}: retrying to connect'
                    report += '\n'
                log(report)
            time.sleep(1)

    except (KeyboardInterrupt, SystemExit):
        sys.tracebacklimit = 0
        for ip in processes:
            processes[ip]['pipe'].send(['EXIT'])

        log('Closing SSH sessions by signal...')
        exit(0)


if __name__ == '__main__':
    main()
